import sys
from thread import get_ident, start_new_thread
from threading import currentThread as current_thread, _active, _active_limbo_lock, _limbo, _MainThread

from mext.log import Logger
from mext.classhacks import get_name
from mext.context import State

from mext.threads.sync import Event#, Condition

__all__ = ['Thread']



class Thread(object):
    state = None
    log = Logger('threads')
    ignore_errors = False
    #_exc_clear = sys.exc_clear

    def __init__(self, func, *args, **kw):
        self.func = func
        self.name = get_name(func, 'Thread-%s' % func)
        self.args = args
        self.kw = kw

        self.daemon = current_thread().daemon
        self.id = None
        self.evt_started = Event()
        self.evt_stopped = Event()

    def __repr__(self):
        if self.evt_stopped.is_set():
            status = "stopped"
        elif self.evt_started.is_set():
            status = "started"
        else:
            status = "initial"
        if self.daemon:
            status += " daemon"
        if self.id is not None:
            status += " %s" % self.id
        return "<%s(%s, %s)>" % (self.__class__.__name__, self.name, status)




    @property
    def is_current(self):
        return self.id == get_ident()

    is_main = False
##     @property
##     def is_main(self):
##         return self.id == main_thread_id
##

    get = get_current = staticmethod(current_thread)

    @classmethod
    def find(self, id):
        return _active.get(id)






    def start_daemon(self):
        self.daemon = True
        return self.start()

    def start(self):
        if self.state is None:
            self.state = State.clone()
        #self.log.info("Starting new thread %s", self.name)
        if self.evt_started.is_set():
            raise RuntimeError("thread already started")
        with _active_limbo_lock:
            _limbo[self] = self
        try:
            start_new_thread(self._thread_main, ())
        except:
            with _active_limbo_lock:
                del _limbo[self]
            raise
        self.evt_started.wait()
        return self



    def _thread_main(self):
        try:
            try:
                self.id = get_ident()
                self.evt_started.set()
                with _active_limbo_lock:
                    _active[self.id] = self
                    del _limbo[self]
                self._thread_main_inner()
            finally:
                with _active_limbo_lock:
                    self.evt_stopped.set()
                    _active.pop(get_ident(), None)
        except:
            if not self.daemon or sys is not None:
                raise


    def _thread_main_inner(self):
        with self.state:
            try:
                #self.log.info("Thread %s is running", self.name)
                self.func(*self.args, **self.kw)
            except SystemExit:
                raise
            except:
                self._handle_error()
            finally:
                #self.log.debug("Thread stopped")
                # Avoid a refcycle if the thread is running a function with
                # an argument that has a member that points to the thread.
                del self.func, self.args, self.kw
                #self._exc_clear()

    def _handle_error(self):
        if sys is None:
            raise
        exc_info = sys.exc_info()
        if not self.ignore_errors:
            sys.excepthook(*exc_info)
        #elif isinstance(self.ignore_errors, Logger)::
        elif self.ignore_errors == 'log':
            self.log.error("Thread raised an exception", exc_info=True)
        else:
            self.log.debug(u"Thread raised an exception %s", exc_info[1])
        del exc_info

    def join(self, timeout=None):
        if not self.evt_started.is_set():
            raise RuntimeError("cannot join thread before it is started")
        if self.is_current:
            raise RuntimeError("cannot join current thread")
        self.evt_stopped.wait(timeout)



    # this makes threading._after_fork work
    def __stopped(self, v):
        self.evt_stopped.set()
    __stopped = property(fset=__stopped)

    # required for threading._pickSomeNonDaemonThread
    def is_alive(self):
        return self.evt_started.is_set() and not self.evt_stopped.is_set()

    alive = property(is_alive)

    # fix for mext.log
    def getName(self):
        return self.name


_MainThread.is_main = True












## t = Thread(lambda: x)
## t.ignore_errors = True
## t.start()
##
##
## @Thread
## def loop():
##     while True:
##         pass
##
## #loop.e = sys.stderr
## loop.start_daemon()








